
require 'rest-client'
require 'openssl'
require 'securerandom'

module Flow

  class Connection
    attr_reader :url
    attr_reader :access_key
    attr_reader :version

    attr_reader :scheme
    attr_reader :host
    attr_reader :port

    def initialize(api_key, url='https://flow.ribbon.co', version=1)
      raise Flow::Errors::ConnectionError, "API Key must be defined." unless api_key
      raise Flow::Errors::ConnectionError, "URL must be defined." unless url
      raise Flow::Errors::ConnectionError, "Version must be defined." unless version

      @version = version

      @url = (url[-1] == '/' ? url[0..-2] : url).dup.freeze # Remove trailing slash.
      
      if /^([A-Za-z0-9]{20})\.([A-Za-z0-9]{43})$/.match(api_key.to_s)
        @access_key = $1.dup.freeze
        @signing_key = $2.dup.freeze
      else
        raise Flow::Errors::ConnectionError, "Invalid API Key format."
      end

      URI.parse(@url).tap do |uri|
        @scheme = uri.scheme
        @host = uri.host
        @port = uri.port
      end

      # RestClient::Resource.new(
      #   'https://example.com',
      #   :ssl_client_cert  =>  OpenSSL::X509::Certificate.new(File.read("cert.pem")),
      #   :ssl_client_key   =>  OpenSSL::PKey::RSA.new(File.read("key.pem"), "passphrase, if any"),
      #   :ssl_ca_file      =>  "ca_certificate.pem",
      #   :verify_ssl       =>  OpenSSL::SSL::VERIFY_PEER
      # )

      @api = RestClient::Resource.new(
        @url,
        verify_ssl: true,
        headers: {
          "Content-Type" => "application/json"
        }
      ) {|response, request, result| response}
    end

    def get(path, params={})
      path = _build_path(path)
      auth_header = _authorization_header(:get, path, params)
      response = @api[path].get(authorization: auth_header, params: params)
      Flow::Response.new(self, response)
    end

    def post(path, params={}, &block)
      if block_given?
        body_hash = block.call
        if body_hash && !body_hash.empty?
          body = body_hash.to_json
        end
      end

      path = _build_path(path)
      auth_header = _authorization_header(:post, path, params, body)
      response = @api[path].post(body, authorization: auth_header, params: params)
      
      Flow::Response.new(self, response)
    end

    private
    
    def _signing_key
      @signing_key
    end

    def _build_path(path)
      path = '/' + path unless path[0] == '/'

      if path.start_with?('/api/')
        return path
      else
        return "/api/v#{version}#{path}"
      end
    end

    def _gen_nonce
      begin
        nonce = SecureRandom.hex(8)
      end until _valid_nonce?(nonce)

      return nonce.freeze
    end

    def _valid_nonce?(nonce)
      record = @__nonce_record ||= {}
      nonce = nonce.to_s

      # Try to limit the record store to about 2MB
      # of data, assuming 16 single-byte characters 
      # and 4 byte timestamps.
      # Supports 3,333 requests per second before exceeding
      # the limit.
      if record.length > 100_000
        threshold_ts = (Time.now.to_i - 30).freeze
        record.delete_if {|k,v| v < threshold_ts}
      end

      if record[nonce] && record[nonce] >= (Time.now.to_i - 30)
        return false
      else
        record[nonce] = Time.now.to_i.freeze
        return true
      end
    end

    def _authorization_header(verb, path, query_hash, body=nil)
      timestamp = Time.now.to_i.freeze
      nonce = _gen_nonce

      # Calculate request hash
      path = _uri_encode(path[0] == '/' ? path : ('/' + path), false)
      query_string = _hash_to_query_string(query_hash)
      body_hash = _sha256(body.to_s)
      request_hash = _sha256(
        verb.to_s.upcase + "\n" +
        host + "\n" +
        path + "\n" +
        query_string + "\n" +
        body_hash
      )

      # Construct the message
      message = timestamp.to_s + nonce + request_hash

      signature = _hmac_sha256_hex(_signing_key, message)

      return "RBN1-HMAC-SHA256 #{access_key};#{timestamp};#{nonce};#{signature}"
    end

    ######################################################################
    #
    # Encoding and signing helper methods
    #
    ######################################################################

    def _hmac_sha256_hex(key, message)
      digest = OpenSSL::Digest::SHA256.new
      OpenSSL::HMAC.hexdigest(digest, key, message)
    end

    def _sha256(str)
      ::OpenSSL::Digest::SHA256.hexdigest(str)
    end

    def _uri_encode(string, encode_fslash=true)
      retval = ""

      string.chars.each do |ch|
        if (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') ||
           (ch >= '0' && ch <= '9') || ch == '-' || ch == '_' || ch == '.' || ch == '~'
          retval << ch
        elsif ch == '/'
          retval << (encode_fslash ? '%2F' : ch)
        else
          retval << '%' << ch.bytes.first.to_s(16).upcase
        end
      end
      
      return retval
    end

    def _hash_to_query_string(data)
      pairs = []

      data.sort{|a,b| a[0] <=> b[0]}.each do |key, value|
        pairs << (_uri_encode(key.to_s) << '=' << _uri_encode(value.to_s))
      end

      pairs.join('&')
    end
  end

end